# -*- coding: UTF8 -*-
"""
container class for results
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import numpy as np


class Result(object):
    def __init__(self):
        self.info = {}
        self.stats = {}
        self.np_arrays = {}
        self.trajectories = {}

    def __str__(self):
        return self.pretty_str(stats=True)

    def __eq__(self, other):
        if not isinstance(other, Result):
            return False
        equal = (self.info == other.info)
        equal &= (self.stats == other.stats)
        equal &= (self.trajectories == other.trajectories)
        for k in self.np_arrays:
            if k not in other.np_arrays:
                equal &= False
                break
            if not equal:
                break
            equal &= all(
                [np.array_equal(self.np_arrays[k], other.np_arrays[k])])
        return equal

    def __ne__(self, other):
        return not self == other

    def pretty_str(self, title=True, stats=True, info=False):
        p_str = ""
        p_str += "{}\n\n".format(self.info["title"]) if title else ""
        if stats:
            for name, val in sorted(self.stats.items()):
                p_str += "{:>10}\t{:.6f}\n".format(name, val)
        if info:
            for name, val in sorted(self.info.items()):
                p_str += "{:>10}\t{}\n".format(name, val)
        return p_str

    def add_np_array(self, name, array):
        self.np_arrays[name] = array

    def add_info(self, info_dict):
        self.info.update(info_dict)

    def add_stats(self, stats_dict):
        self.stats.update(stats_dict)

    def add_trajectory(self, name, traj):
        self.trajectories[name] = traj



# https://docs.python.org/3/howto/logging.html#library-config

import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())



#!/usr/bin/env python

from evo import main_evaluation as me

if __name__ == '__main__':

    if True:
        test_name = "smartNoiseSigma"
        extra_params_to_modify = dict()
        values_to_test = [1.0, 1.25, 1.5, 1.75, 2.0, 2.25, 2.5, 2.75, 3, 3.25, 3.5]
        me.regression_test_simple(test_name, ["smartNoiseSigma"], values_to_test,
                                  False, True, 1, extra_params_to_modify) # 1 means only run S

#BSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
    if False:
        test_name = "Mono_Stereo_L2"
        extra_params_to_modify['monoNormType'] = 0
        extra_params_to_modify['stereoNormType'] = 0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Huber_0.5"
        extra_params_to_modify['monoNormType'] = 1
        extra_params_to_modify['monoNormParam'] = 0.5
        extra_params_to_modify['stereoNormType'] = 1
        extra_params_to_modify['stereoNormParam'] = 0.5
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Huber_1.0"
        extra_params_to_modify['monoNormType'] = 1
        extra_params_to_modify['monoNormParam'] = 1.0
        extra_params_to_modify['stereoNormType'] = 1
        extra_params_to_modify['stereoNormParam'] = 1.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

        test_name = "Mono_Stereo_Huber_1.5"
        extra_params_to_modify['monoNormType'] = 1
        extra_params_to_modify['monoNormParam'] = 1.5
        extra_params_to_modify['stereoNormType'] = 1
        extra_params_to_modify['stereoNormParam'] = 1.5
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

        test_name = "Mono_Stereo_Huber_2.0"
        extra_params_to_modify['monoNormType'] = 1
        extra_params_to_modify['monoNormParam'] = 2.0
        extra_params_to_modify['stereoNormType'] = 1
        extra_params_to_modify['stereoNormParam'] = 2.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Tukey_1.0"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 1.0
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 1.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

        test_name = "Mono_Stereo_Tukey_1.5"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 1.5
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 1.5
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

        test_name = "Mono_Stereo_Tukey_2.0"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 2.0
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 2.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Tukey_3.0"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 3.0
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 3.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Tukey_4.0"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 4.0
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 4.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Tukey_4.6851"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 4.6851
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 4.6851
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

    if False:
        test_name = "Mono_Stereo_Tukey_5.0"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 5.0
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 5.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR

        test_name = "Mono_Stereo_Tukey_6.0"
        extra_params_to_modify['monoNormType'] = 2
        extra_params_to_modify['monoNormParam'] = 6.0
        extra_params_to_modify['stereoNormType'] = 2
        extra_params_to_modify['stereoNormParam'] = 6.0
        params_to_test = ["monoNoiseSigma", "stereoNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [[0.8, 0.8],
                                                              [1.0, 1.0],
                                                              [1.2, 1.2],
                                                              [1.4, 1.4],
                                                              [1.6, 1.6],
                                                              [1.8, 1.8]],
                                  True, True, 6, extra_params_to_modify) # 6 means run SP and SPR
        ################## Smart Noise #######################################################

    if False:
        test_name = "Smart"
        params_to_test = ["smartNoiseSigma"]
        me.regression_test_simple(test_name, params_to_test, [1.0, 1.25, 1.5, 1.75, 2.0, 2.25, 2.5,
                                                              2.75, 3.0, 3.25, 3.5],
                                  True, True, 0, extra_params_to_modify) # 0 means run S, SP and SPR


        ################## REG NORM #######################################################

    if False:
        test_name = "Tukey_4.0"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 4.0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_4.3"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 4.3
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_4.6851"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 4.6851
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_5.0"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 5.0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_1.0"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 1.0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_2.0"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 2.0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_3.0"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 3.0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Tukey_3.5"
        extra_params_to_modify['regularityNormType'] = 2
        extra_params_to_modify['regularityNormParam'] = 3.5
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR


    if False:
        test_name = "L2"
        extra_params_to_modify['regularityNormType'] = 0
        extra_params_to_modify['regularityNormParam'] = 0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

    if False:
        test_name = "Huber_0.5"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 0.5
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Huber_0.6"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 0.6
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

    if False:
        test_name = "Huber_0.7"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 0.7
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Huber_0.8"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 0.8
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Huber_0.9"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 0.9
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Huber_1.0"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 1.0
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

        test_name = "Huber_1.1"
        extra_params_to_modify['regularityNormType'] = 1
        extra_params_to_modify['regularityNormParam'] = 1.1
        me.regression_test_simple(test_name, ["regularityNoiseSigma"], [0.02, 0.03, 0.04, 0.05], True, True, 3, extra_params_to_modify) # 3 means only run SPR

    #me.regression_test_simple("stereoNoiseSigma", [1.5, 1.6, 1.7])
    #me.regression_test_simple("monoNoiseSigma", [1.5, 1.6, 1.7])



# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
separate entry points into pieces to allow common error handling and faster argcomplete
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import logging

import argcomplete

logger = logging.getLogger(__name__)
""" 
the actual entry points:
to save time for argcomplete (tab bash completion),
only do required imports in respective module when creating parser
(no expensive global imports)
"""

def evaluation():
    from evo import main_evaluation
    parser = main_evaluation.parser()
    argcomplete.autocomplete(parser)
    launch(main_evaluation, parser)

def ape():
    from evo import main_ape
    parser = main_ape.parser()
    argcomplete.autocomplete(parser)
    launch(main_ape, parser)


def rpe():
    from evo import main_rpe
    parser = main_rpe.parser()
    argcomplete.autocomplete(parser)
    launch(main_rpe, parser)


def res():
    from evo import main_res
    parser = main_res.parser()
    argcomplete.autocomplete(parser)
    launch(main_res, parser)


def traj():
    from evo import main_traj
    parser = main_traj.parser()
    argcomplete.autocomplete(parser)
    launch(main_traj, parser)


def merge_config(args):
    """
    merge .json config file with the command line args (if --config was defined)
    :param args: parsed argparse NameSpace object
    :return: merged argparse NameSpace object
    """
    import json
    if args.config:
        with open(args.config) as config:
            merged_config_dict = vars(args).copy()
            # merge both parameter dicts
            merged_config_dict.update(json.loads(config.read()))
            # override args the hacky way
            args = argparse.Namespace(**merged_config_dict)
    return args


def launch(main_module, parser):
    args = parser.parse_args()
    if hasattr(args, "config"):
        args = merge_config(args)
    import sys
    from evo.tools import settings
    try:
        main_module.run(args)
    except SystemExit as e:
        sys.exit(e.code)
    except:
        logger.exception("Unhandled error in " + main_module.__name__)
        print("")
        err_msg = "evo module " + main_module.__name__ + " crashed"
        if settings.SETTINGS.logfile_enabled:
            err_msg += " - see " + settings.DEFAULT_LOGFILE_PATH
        else:
            err_msg += " - no logfile written (disabled)"
        logger.error(err_msg)
        from evo.tools import user
        if not args.no_warnings:
            if settings.SETTINGS.logfile_enabled and user.confirm(
                    "Open logfile? (y/n)"):
                import webbrowser
                webbrowser.open(settings.DEFAULT_LOGFILE_PATH)
        sys.exit(1)



#!/usr/bin/env python
"""
unit test for filters module
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import math
import unittest

import numpy as np

import context
from evo.core import filters
from evo.core import lie_algebra as lie

poses_1 = [
    lie.se3(np.eye(3), np.array([0, 0, 0])),
    lie.se3(np.eye(3), np.array([0, 0, 0.5])),
    lie.se3(np.eye(3), np.array([0, 0, 0])),
    lie.se3(np.eye(3), np.array([0, 0, 1]))
]

poses_2 = [
    lie.se3(np.eye(3), np.array([0, 0, 0])),
    lie.se3(np.eye(3), np.array([0, 0, 0.5])),
    lie.se3(np.eye(3), np.array([0, 0, 0.99])),
    lie.se3(np.eye(3), np.array([0, 0, 1.0]))
]

poses_3 = [
    lie.se3(np.eye(3), np.array([0, 0, 0.0])),
    lie.se3(np.eye(3), np.array([0, 0, 0.9])),
    lie.se3(np.eye(3), np.array([0, 0, 0.99])),
    lie.se3(np.eye(3), np.array([0, 0, 0.999])),
    lie.se3(np.eye(3), np.array([0, 0, 0.9999])),
    lie.se3(np.eye(3), np.array([0, 0, 0.99999])),
    lie.se3(np.eye(3), np.array([0, 0, 0.999999])),
    lie.se3(np.eye(3), np.array([0, 0, 0.9999999]))
]

poses_4 = [
    lie.se3(np.eye(3), np.array([0, 0, 0])),
    lie.se3(np.eye(3), np.array([0, 0, 1])),
    lie.se3(np.eye(3), np.array([0, 0, 1])),
    lie.se3(np.eye(3), np.array([0, 0, 1]))
]


class TestFilterPairsByPath(unittest.TestCase):
    def test_poses1_all_pairs(self):
        target_path = 1.0
        tol = 0.0
        id_pairs = filters.filter_pairs_by_path(poses_1, target_path, tol,
                                                all_pairs=True)
        self.assertEqual(id_pairs, [(0, 2), (2, 3)])

    def test_poses1_wrong_target(self):
        target_path = 2.5
        tol = 0.0
        id_pairs = filters.filter_pairs_by_path(poses_1, target_path, tol,
                                                all_pairs=True)
        self.assertEqual(id_pairs, [])

    def test_poses2_all_pairs_low_tolerance(self):
        target_path = 1.0
        tol = 0.001
        id_pairs = filters.filter_pairs_by_path(poses_2, target_path, tol,
                                                all_pairs=True)
        self.assertEqual(id_pairs, [(0, 3)])

    def test_convergence_all_pairs(self):
        target_path = 1.0
        tol = 0.2
        id_pairs = filters.filter_pairs_by_path(poses_3, target_path, tol,
                                                all_pairs=True)
        self.assertEqual(id_pairs, [(0, 7)])


class TestFilterPairsByDistance(unittest.TestCase):
    def test_poses1_all_pairs(self):
        target_path = 1.0
        tol = 0.0
        id_pairs = filters.filter_pairs_by_distance(poses_1, target_path, tol,
                                                    all_pairs=True)
        self.assertEqual(id_pairs, [(0, 3), (2, 3)])

    def test_poses1_wrong_target(self):
        target_path = 2.5
        tol = 0.0
        id_pairs = filters.filter_pairs_by_distance(poses_1, target_path, tol,
                                                    all_pairs=True)
        self.assertEqual(id_pairs, [])

    def test_poses2_all_pairs_low_tolerance(self):
        target_path = 1.0
        tol = 0.001
        id_pairs = filters.filter_pairs_by_distance(poses_2, target_path, tol,
                                                    all_pairs=True)
        self.assertEqual(id_pairs, [(0, 3)])

    def test_poses4_all_pairs(self):
        target_path = 1.0
        tol = 0.2
        id_pairs = filters.filter_pairs_by_distance(poses_4, target_path, tol,
                                                    all_pairs=True)
        self.assertEqual(id_pairs, [(0, 1), (0, 2), (0, 3)])


axis = np.array([1, 0, 0])
poses_5 = [
    lie.se3(lie.so3_exp(axis, 0.0), np.array([0, 0, 0])),
    lie.se3(lie.so3_exp(axis, math.pi), np.array([0, 0, 0])),
    lie.se3(lie.so3_exp(axis, 0.0), np.array([0, 0, 0])),
    lie.se3(lie.so3_exp(axis, math.pi / 3), np.array([0, 0, 0])),
    lie.se3(lie.so3_exp(axis, math.pi), np.array([0, 0, 0]))
]
stamps_5 = np.array([0, 1, 2, 3, 4])

axis = np.array([1, 0, 0])
p0 = lie.se3(lie.so3_exp(axis, 0.0), np.array([0, 0, 0]))
pd = lie.se3(lie.so3_exp(axis, math.pi / 3), np.array([1, 2, 3]))
p1 = np.dot(p0, pd)
p2 = np.dot(p1, pd)
p3 = np.dot(p2, pd)
poses_6 = [p0, p1, p2, p3, p3]
stamps_6 = np.array([0, 1, 2, 3, 4])


class TestFilterPairsByAngle(unittest.TestCase):
    def test_poses5(self):
        target_angle = math.pi
        tol = 0.001
        id_pairs = filters.filter_pairs_by_angle(poses_5, target_angle, tol,
                                                 all_pairs=False)
        self.assertEqual(id_pairs, [(0, 1), (1, 2), (2, 4)])

    def test_poses5_all_pairs(self):
        target_angle = math.pi
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angle(poses_5, target_angle, tol,
                                                 all_pairs=True)
        self.assertEqual(id_pairs, [(0, 1), (0, 4), (1, 2), (2, 4)])

    def test_poses6(self):
        target_angle = math.pi
        tol = 0.001
        id_pairs = filters.filter_pairs_by_angle(poses_6, target_angle, tol,
                                                 all_pairs=False)
        self.assertEqual(id_pairs, [(0, 3)])

    def test_poses6_all_pairs(self):
        target_angle = math.pi
        tol = 0.001
        id_pairs = filters.filter_pairs_by_angle(poses_6, target_angle, tol,
                                                 all_pairs=True)
        self.assertEqual(id_pairs, [(0, 3), (0, 4)])


poses_7 = [
    lie.se3(np.eye(3), np.array([0, 0, 1])),
    lie.se3(np.eye(3), np.array([0, 0, 2])),
    lie.se3(np.eye(3), np.array([0, 0, 3])),
    lie.se3(np.eye(3), np.array([0, 0, 4]))
]
stamps_7 = np.array([0, 1, 2, 3])

poses_8 = [
    lie.se3(np.eye(3), np.array([0, 0, 1])),
    lie.se3(np.eye(3), np.array([0, 0, 1])),
    lie.se3(np.eye(3), np.array([0, 0, 3])),
    lie.se3(np.eye(3), np.array([0, 0, 4]))
]
stamps_8 = np.array([0, 1, 2, 3])


class TestFilterPosesBySpeed(unittest.TestCase):
    def test_poses_7_stamps_7(self):
        target_speed = 1.0  # m/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_speed(poses_7, stamps_7,
                                                 target_speed, tol)
        self.assertEqual(id_pairs, [(0, 1), (1, 2), (2, 3)])

    def test_poses_8_stamps_8(self):
        target_speed = 1.0  # m/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_speed(poses_8, stamps_8,
                                                 target_speed, tol)
        self.assertEqual(id_pairs, [(2, 3)])


class TestFilterPosesByAngularSpeed(unittest.TestCase):
    def test_poses_7_stamps_7_radians(self):
        target_speed = math.pi / 3  # rad/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angular_speed(
            poses_7, stamps_7, target_speed, tol)
        self.assertEqual(id_pairs, [])

    def test_poses_6_stamps_6_radians(self):
        target_speed = math.pi / 3  # rad/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angular_speed(
            poses_6, stamps_6, target_speed, tol)
        self.assertEqual(id_pairs, [(0, 1), (1, 2), (2, 3)])

    def test_poses_5_stamps_5_radians(self):
        target_speed = math.pi / 3  # rad/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angular_speed(
            poses_5, stamps_5, target_speed, tol)
        self.assertEqual(id_pairs, [(2, 3)])

    def test_poses_7_stamps_7_degrees(self):
        target_speed = math.pi / 3 * 180 / math.pi  # deg/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angular_speed(
            poses_7, stamps_7, target_speed, tol, degrees=True)
        self.assertEqual(id_pairs, [])

    def test_poses_6_stamps_6_degrees(self):
        target_speed = math.pi / 3.0 * 180 / math.pi  # rad/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angular_speed(
            poses_6, stamps_6, target_speed, tol, degrees=True)
        self.assertEqual(id_pairs, [(0, 1), (1, 2), (2, 3)])

    def test_poses_5_stamps_5_degrees(self):
        target_speed = math.pi / 3 * 180 / math.pi  # rad/s
        tol = 0.01
        id_pairs = filters.filter_pairs_by_angular_speed(
            poses_5, stamps_5, target_speed, tol, degrees=True)
        self.assertEqual(id_pairs, [(2, 3)])


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""
Provides functionality for loading and resetting the package settings.
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import os
import json
import logging

from colorama import Fore

logger = logging.getLogger(__name__)

PACKAGE_BASE_PATH = os.path.abspath(__file__ + "/../../")
PACKAGE_VERSION = open(os.path.join(PACKAGE_BASE_PATH, "version")).read()
USER_ASSETS_PATH = os.path.join(os.path.expanduser('~'), ".evo")
USER_ASSETS_VERSION_PATH = os.path.join(USER_ASSETS_PATH, "assets_version")
DEFAULT_PATH = os.path.join(USER_ASSETS_PATH, "settings.json")
DEFAULT_LOGFILE_PATH = os.path.join(USER_ASSETS_PATH, "evo.log")


class SettingsException(Exception):
    pass


class SettingsContainer(dict):
    def __init__(self, data, lock=True):
        super(SettingsContainer, self).__init__()
        for k, v in data.items():
            setattr(self, k, v)
        setattr(self, "__locked__", lock)

    @classmethod
    def from_json_file(cls, settings_path):
        with open(settings_path) as settings_file:
            data = json.load(settings_file)
        return SettingsContainer(data)

    def locked(self):
        if "__locked__" in self:
            return self["__locked__"]

    def __getattr__(self, attr):
        # allow dot access
        if attr not in self:
            raise SettingsException("unknown settings parameter: " + str(attr))
        return self[attr]

    def __setattr__(self, attr, value):
        # allow dot access
        if self.locked() and attr not in self:
            raise SettingsException(
                "write-access locked, can't add new parameter {}".format(attr))
        else:
            self[attr] = value


def merge_dicts(first, second, soft=False):
    if soft:
        first.update({k: v for k, v in second.items() if k not in first})
    else:
        first.update(second)
    return first


def write_to_json_file(json_path, dictionary):
    with open(json_path, 'w') as json_file:
        json_file.write(json.dumps(dictionary, indent=4, sort_keys=True))


def reset(dest=DEFAULT_PATH):
    from evo.tools.settings_template import DEFAULT_SETTINGS_DICT
    write_to_json_file(dest, DEFAULT_SETTINGS_DICT)


def initialize_if_needed():
    """
    Initialize evo user folder after first installation
    (or if it was deleted).
    """
    if not os.path.isdir(USER_ASSETS_PATH):
        os.makedirs(USER_ASSETS_PATH)

    if not os.path.exists(USER_ASSETS_VERSION_PATH):
        open(USER_ASSETS_VERSION_PATH, 'w').write(PACKAGE_VERSION)

    if not os.path.exists(DEFAULT_PATH):
        try:
            reset(dest=DEFAULT_PATH)
            print("{}Initialized new {}{}".format(Fore.LIGHTYELLOW_EX,
                                                  DEFAULT_PATH, Fore.RESET))
        except:
            logger.error("Fatal: failed to write package settings file {}".
                         format(DEFAULT_PATH))
            raise


def update_if_outdated():
    """
    Update user settings to a new version if needed.
    """
    if open(USER_ASSETS_VERSION_PATH).read() == PACKAGE_VERSION:
        return
    from evo.tools.settings_template import DEFAULT_SETTINGS_DICT
    old_settings = json.loads(open(DEFAULT_PATH).read())
    updated_settings = merge_dicts(old_settings, DEFAULT_SETTINGS_DICT,
                                   soft=True)
    write_to_json_file(DEFAULT_PATH, updated_settings)
    open(USER_ASSETS_VERSION_PATH, 'w').write(PACKAGE_VERSION)
    print("{}Updated outdated {}{}".format(Fore.LIGHTYELLOW_EX, DEFAULT_PATH,
                                           Fore.RESET))


# Load the user settings into this container.
initialize_if_needed()
update_if_outdated()
SETTINGS = SettingsContainer.from_json_file(DEFAULT_PATH)



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
launch a custom IPython shell for evo
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import os
import six
import sys
import shutil
import argparse
import subprocess as sp

from evo.tools.settings import PACKAGE_BASE_PATH
from evo.tools.compat import which

DESC = '''
Launches an IPython shell with pre-loaded evo modules
(c) michael.grupp@tum.de

Unknown command line arguments are forwarded to the ipython executable
'''


def main():
    main_parser = argparse.ArgumentParser(
        description=DESC, formatter_class=argparse.RawTextHelpFormatter)
    args, other_args = main_parser.parse_known_args()
    other_args = [] if other_args is None else other_args
    FNULL = open(os.devnull, 'w')

    # check if IPython is installed properly
    ipython = "ipython3" if six.PY3 else "ipython2"
    if which(ipython) is None:
        # fall back to the non-explicit ipython name if ipython2/3 is not in PATH
        ipython = "ipython"
        if which(ipython) is None:
            print("IPython is not installed", file=sys.stderr)
            sys.exit(1)

    try:
        sp.check_call([ipython, "profile", "locate", "evo"], stdout=FNULL,
                      stderr=FNULL)
    except sp.CalledProcessError:
        print("IPython profile for evo is not installed", file=sys.stderr)
        sp.call([ipython, "profile", "create", "evo"])
        config = os.path.join(PACKAGE_BASE_PATH, "ipython_config.py")
        profile_dir = sp.check_output([ipython, "profile", "locate",
                                       "evo"]).decode("utf-8")
        profile_dir = profile_dir.rstrip()
        shutil.copy(config, os.path.join(profile_dir, "ipython_config.py"))
    try:
        sp.check_call([ipython, "--profile", "evo"] + other_args)
    except sp.CalledProcessError as e:
        print("IPython error", e.output, file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main executable for viewing result files from the trajectory metric apps
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import logging

logger = logging.getLogger(__name__)

SEP = "-" * 80  # separator line

CONFLICT_TEMPLATE = """
Mismatching titles - risk of aggregating data from different metrics. Conflict:

<<<<<<< {0}
{1}
=======
{2}
>>>>>>> {3}

Only the first one will be used as the title!"""


def parser():
    import argparse
    basic_desc = "tool for processing one or multiple result files"
    lic = "(c) michael.grupp@tum.de"
    main_parser = argparse.ArgumentParser(
        description="%s %s" % (basic_desc, lic))
    output_opts = main_parser.add_argument_group("output options")
    usability_opts = main_parser.add_argument_group("usability options")
    main_parser.add_argument("result_files",
                             help="one or multiple result files", nargs='+')
    main_parser.add_argument("--use_rel_time",
                             help="use relative timestamps if available",
                             action="store_true")
    main_parser.add_argument("--use_filenames",
                             help="use the filenames to label the data",
                             action="store_true")
    output_opts.add_argument("-p", "--plot", help="show plot window",
                             action="store_true")
    output_opts.add_argument("--plot_markers", help="plot with circle markers",
                             action="store_true")
    output_opts.add_argument("--save_plot", help="path to save plot",
                             default=None)
    output_opts.add_argument("--serialize_plot",
                             help="path to serialize plot (experimental)",
                             default=None)
    output_opts.add_argument(
        "--save_table", help="path to a file to save the results in a table",
        default=None)
    usability_opts.add_argument("--no_warnings",
                                help="no warnings requiring user confirmation",
                                action="store_true")
    usability_opts.add_argument("-v", "--verbose", help="verbose output",
                                action="store_true")
    usability_opts.add_argument("--silent", help="don't print any output",
                                action="store_true")
    usability_opts.add_argument(
        "--debug", help="verbose output with additional debug info",
        action="store_true")
    usability_opts.add_argument(
        "-c", "--config",
        help=".json file with parameters (priority over command line args)")
    return main_parser


def run(args):
    import sys

    import pandas as pd

    from evo.tools import file_interface, log, user, settings, pandas_bridge
    from evo.tools.settings import SETTINGS

    pd.options.display.width = 80
    pd.options.display.max_colwidth = 20

    log.configure_logging(args.verbose, args.silent, args.debug)
    if args.debug:
        import pprint
        arg_dict = {arg: getattr(args, arg) for arg in vars(args)}
        logger.debug("main_parser config:\n{}\n".format(
            pprint.pformat(arg_dict)))

    df = pd.DataFrame()
    for result_file in args.result_files:
        result = file_interface.load_res_file(result_file)
        name = result_file if args.use_filenames else None
        df = pd.concat([df, pandas_bridge.result_to_df(result, name)],
                       axis="columns")

    keys = df.columns.values.tolist()
    if SETTINGS.plot_usetex:
        keys = [key.replace("_", "\\_") for key in keys]
        df.columns = keys
    duplicates = [x for x in keys if keys.count(x) > 1]
    if duplicates:
        logger.error("Values of 'est_name' must be unique - duplicates: {}\n"
                     "Try using the --use_filenames option to use filenames "
                     "for labeling instead.".format(", ".join(duplicates)))
        sys.exit(1)

    # derive a common index type if possible - preferably timestamps
    common_index = None
    time_indices = ["timestamps", "seconds_from_start", "sec_from_start"]
    if args.use_rel_time:
        del time_indices[0]
    for idx in time_indices:
        if idx not in df.loc["np_arrays"].index:
            continue
        if df.loc["np_arrays", idx].isnull().values.any():
            continue
        else:
            common_index = idx
            break

    # build error_df (raw values) according to common_index
    if common_index is None:
        # use a non-timestamp index
        error_df = pd.DataFrame(df.loc["np_arrays", "error_array"].tolist(),
                                index=keys).T
    else:
        error_df = pd.DataFrame()
        for key in keys:
            new_error_df = pd.DataFrame({
                key: df.loc["np_arrays", "error_array"][key]
            }, index=df.loc["np_arrays", common_index][key])
            duplicates = new_error_df.index.duplicated(keep="first")
            if any(duplicates):
                logger.warning(
                    "duplicate indices in error array of {} - "
                    "keeping only first occurrence of duplicates".format(key))
                new_error_df = new_error_df[~duplicates]
            error_df = pd.concat([error_df, new_error_df], axis=1)

    # check titles
    first_title = df.loc["info", "title"][0]
    first_file = args.result_files[0]
    if not args.no_warnings:
        checks = df.loc["info", "title"] != first_title
        for i, differs in enumerate(checks):
            if not differs:
                continue
            else:
                mismatching_title = df.loc["info", "title"][i]
                mismatching_file = args.result_files[i]
                logger.debug(SEP)
                logger.warning(
                    CONFLICT_TEMPLATE.format(first_file, first_title,
                                             mismatching_title,
                                             mismatching_file))
                if not user.confirm(
                        "Go on anyway? - enter 'y' or any other key to exit"):
                    sys.exit()

    logger.debug(SEP)
    logger.debug("Aggregated dataframe:\n{}".format(
        df.to_string(line_width=80)))

    # show a statistics overview
    logger.debug(SEP)
    logger.info("\n{}\n\n{}\n".format(
        first_title, df.loc["stats"].T.to_string(line_width=80)))

    if args.save_table:
        logger.debug(SEP)
        if args.no_warnings or user.check_and_confirm_overwrite(
                args.save_table):
            if SETTINGS.table_export_data.lower() == "error_array":
                data = error_df
            elif SETTINGS.table_export_data.lower() in ("info", "stats"):
                data = df.loc[SETTINGS.table_export_data.lower()]
            else:
                raise ValueError(
                    "unsupported export data specifier: {}".format(
                        SETTINGS.table_export_data))
            if SETTINGS.table_export_transpose:
                data = data.T

            if SETTINGS.table_export_format == "excel":
                writer = pd.ExcelWriter(args.save_table)
                data.to_excel(writer)
                writer.save()
                writer.close()
            else:
                getattr(data,
                        "to_" + SETTINGS.table_export_format)(args.save_table)
            logger.debug("{} table saved to: {}".format(
                SETTINGS.table_export_format, args.save_table))

    if args.plot or args.save_plot or args.serialize_plot:
        # check if data has NaN "holes" due to different indices
        inconsistent = error_df.isnull().values.any()
        if inconsistent and common_index != "timestamps" and not args.no_warnings:
            logger.debug(SEP)
            logger.warning("Data lengths/indices are not consistent, "
                           "raw value plot might not be correctly aligned")

        from evo.tools import plot
        import matplotlib.pyplot as plt
        import seaborn as sns
        import math

        # use default plot settings
        figsize = (SETTINGS.plot_figsize[0], SETTINGS.plot_figsize[1])
        use_cmap = SETTINGS.plot_multi_cmap.lower() != "none"
        colormap = SETTINGS.plot_multi_cmap if use_cmap else None
        linestyles = ["-o" for x in args.result_files
                      ] if args.plot_markers else None

        # labels according to first dataset
        title = first_title
        if "xlabel" in df.loc["info"].index and not df.loc[
                "info", "xlabel"].isnull().values.any():
            index_label = df.loc["info", "xlabel"][0]
        else:
            index_label = "$t$ (s)" if common_index else "index"
        metric_label = df.loc["info", "label"][0]

        plot_collection = plot.PlotCollection(title)
        # raw value plot
        fig_raw = plt.figure(figsize=figsize)
        # handle NaNs from concat() above
        error_df.interpolate(method="index").plot(
            ax=fig_raw.gca(), colormap=colormap, style=linestyles,
            title=first_title)
        plt.xlabel(index_label)
        plt.ylabel(metric_label)
        plt.legend(frameon=True)
        plot_collection.add_figure("raw", fig_raw)

        # statistics plot
        fig_stats = plt.figure(figsize=figsize)
        exclude = df.loc["stats"].index.isin(["sse"])  # don't plot sse
        df.loc["stats"][~exclude].plot(kind="barh", ax=fig_stats.gca(),
                                       colormap=colormap, stacked=False)
        plt.xlabel(metric_label)
        plt.legend(frameon=True)
        plot_collection.add_figure("stats", fig_stats)

        # grid of distribution plots
        raw_tidy = pd.melt(error_df, value_vars=list(error_df.columns.values),
                           var_name="estimate", value_name=metric_label)
        col_wrap = 2 if len(args.result_files) <= 2 else math.ceil(
            len(args.result_files) / 2.0)
        dist_grid = sns.FacetGrid(raw_tidy, col="estimate", col_wrap=col_wrap)
        dist_grid.map(sns.distplot, metric_label)  # fits=stats.gamma
        plot_collection.add_figure("histogram", dist_grid.fig)

        # box plot
        fig_box = plt.figure(figsize=figsize)
        ax = sns.boxplot(x=raw_tidy["estimate"], y=raw_tidy[metric_label],
                         ax=fig_box.gca())
        # ax.set_xticklabels(labels=[item.get_text() for item in ax.get_xticklabels()], rotation=30)
        plot_collection.add_figure("box_plot", fig_box)

        # violin plot
        fig_violin = plt.figure(figsize=figsize)
        ax = sns.violinplot(x=raw_tidy["estimate"], y=raw_tidy[metric_label],
                            ax=fig_violin.gca())
        # ax.set_xticklabels(labels=[item.get_text() for item in ax.get_xticklabels()], rotation=30)
        plot_collection.add_figure("violin_histogram", fig_violin)

        if args.plot:
            plot_collection.show()
        if args.save_plot:
            logger.debug(SEP)
            plot_collection.export(args.save_plot,
                                   confirm_overwrite=not args.no_warnings)
        if args.serialize_plot:
            logger.debug(SEP)
            plot_collection.serialize(args.serialize_plot,
                                      confirm_overwrite=not args.no_warnings)


if __name__ == '__main__':
    from evo import entry_points
    entry_points.res()



"""
user interaction functions
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import logging

logger = logging.getLogger(__name__)

# Python 2/3 compatibility
try:
    input = raw_input
except NameError:
    pass


def prompt_val(msg="enter a value:"):
    return input(msg + "\n")


def confirm(msg="enter 'y' to confirm or any other key to cancel", key='y'):
    if input(msg + "\n") != key:
        return False
    else:
        return True


def check_and_confirm_overwrite(file_path):
    if os.path.isfile(file_path):
        logger.warning(file_path + " exists, overwrite?")
        return confirm("enter 'y' to overwrite or any other key to cancel")
    else:
        return True



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
plot editor
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import logging

logger = logging.getLogger(__name__)

SEP = "-" * 80  # separator line


def main():
    import argparse
    import argcomplete
    basic_desc = "experimental tool for opening a serialized PlotCollection"
    lic = "(c) michael.grupp@tum.de"
    main_parser = argparse.ArgumentParser(
        description="%s %s" % (basic_desc, lic))
    main_parser.add_argument("in_file",
                             help="path to a serialized plot_collection")
    main_parser.add_argument("-t", "--title",
                             help="custom title (default: file name)")
    main_parser.add_argument("--save_plot", help="path to save plot",
                             default=None)
    main_parser.add_argument("--serialize_plot",
                             help="path to re-serialize PlotCollection",
                             default=None)
    main_parser.add_argument("--to_html",
                             help="convert to html (requires mpld3 library)",
                             action="store_true")
    main_parser.add_argument("--no_warnings",
                             help="no warnings requiring user confirmation",
                             action="store_true")
    argcomplete.autocomplete(main_parser)
    args = main_parser.parse_args()

    from evo.tools import log, plot, user
    log.configure_logging(verbose=True)

    if not args.title:
        title = os.path.basename(args.in_file)
    else:
        title = args.title
    if not args.no_warnings:
        logger.warning(
            "This tool is experimental and not guranteed to work.\nOnly works "
            "if the same plot settings are used as for serialization.\n"
            "If not, try: evo_config show/set \n" + SEP)

    plot_collection = plot.PlotCollection(title, deserialize=args.in_file)
    logger.debug("Deserialized PlotCollection: " + str(plot_collection))
    plot_collection.show()

    if args.serialize_plot:
        logger.debug(SEP)
        plot_collection.serialize(args.serialize_plot,
                                  confirm_overwrite=not args.no_warnings)
    if args.save_plot:
        logger.debug(SEP)
        plot_collection.export(args.save_plot,
                               confirm_overwrite=not args.no_warnings)
    if args.to_html:
        import mpld3
        logger.debug(SEP + "\nhtml export\n")
        for name, fig in plot_collection.figures.items():
            html = mpld3.fig_to_html(fig)
            out = name + ".html"
            with open(out, 'w') as f:
                logger.debug(out)
                f.write(html)
    if not args.no_warnings:
        logger.debug(SEP)
        if user.confirm("Save changes & overwrite original file " +
                        args.in_file + "? (y/n)"):
            plot_collection.serialize(args.in_file, confirm_overwrite=False)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
unit test for trajectory module
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function  # Python 2.7 backwards compatibility

import os
import unittest
import copy

import numpy as np

import context
import helpers
from evo.core import trajectory, geometry
from evo.core import lie_algebra as lie
from evo.tools import file_interface

# use absolute paths!
here = os.path.dirname(os.path.abspath(__file__))
tum_traj_file = os.path.join(here, "data/fr2_desk_ORB.txt")
kitti_traj_file = os.path.join(here, "data/KITTI_00_gt.txt")

ex_tum_traj = file_interface.read_tum_trajectory_file(tum_traj_file)
ex_kitti_traj = file_interface.read_kitti_poses_file(kitti_traj_file)
ex_kitti_traj_wrong = copy.deepcopy(ex_kitti_traj)
ex_kitti_traj_wrong._poses_se3 = [
    np.zeros((4, 4)) for i in range(ex_kitti_traj.num_poses)
]

ex_tum_traj_wrong_quat = copy.deepcopy(ex_tum_traj)
ex_tum_traj_wrong_quat._orientations_quat_wxyz[3] = [5000, 0, 0, 0]

ex_tum_traj_wrong_stamps = copy.deepcopy(ex_tum_traj)
# unsorted timestamps
ex_tum_traj_wrong_stamps.timestamps = np.roll(
    helpers.fake_timestamps(ex_tum_traj.num_poses, 0.1), 1)


class TestPosePath3D(unittest.TestCase):
    def test_init_wrong_args(self):
        # no args
        with self.assertRaises(trajectory.TrajectoryException):
            trajectory.PosePath3D()
        # only quaternion
        with self.assertRaises(trajectory.TrajectoryException):
            trajectory.PosePath3D(
                orientations_quat_wxyz=ex_tum_traj.orientations_quat_wxyz)
        # only xyz
        with self.assertRaises(trajectory.TrajectoryException):
            trajectory.PosePath3D(positions_xyz=ex_tum_traj.positions_xyz)

    def test_init_correct(self):
        # only poses_se3
        try:
            trajectory.PosePath3D(poses_se3=ex_tum_traj.poses_se3)
        except trajectory.TrajectoryException:
            self.fail("unexpected init failure with only poses_se3")
        # xyz + quaternion
        try:
            trajectory.PosePath3D(ex_tum_traj.positions_xyz,
                                  ex_tum_traj.orientations_quat_wxyz)
        except trajectory.TrajectoryException:
            self.fail("unexpected init failure with xyz + quaternion")
        # all
        try:
            trajectory.PosePath3D(ex_tum_traj.positions_xyz,
                                  ex_tum_traj.orientations_quat_wxyz,
                                  ex_tum_traj.poses_se3)
        except trajectory.TrajectoryException:
            self.fail(
                "unexpected init failure with xyz + quaternion + poses_se3")

    def test_equals(self):
        ex_kitti_traj_copy = copy.deepcopy(ex_kitti_traj)
        self.assertTrue(ex_kitti_traj == ex_kitti_traj_copy)
        self.assertFalse(ex_kitti_traj == ex_kitti_traj_wrong)
        self.assertTrue(ex_kitti_traj != ex_kitti_traj_wrong)
        self.assertFalse(ex_kitti_traj != ex_kitti_traj_copy)

    def test_reduce_to_ids(self):
        traj_reduced = copy.deepcopy(ex_kitti_traj)
        traj_reduced.reduce_to_ids([0, 2])
        self.assertEqual(traj_reduced.num_poses, 2)
        # direct connection from 0 to 2 in initial should be reduced path length
        len_initial_segment = np.linalg.norm(ex_kitti_traj.positions_xyz[2] -
                                             ex_kitti_traj.positions_xyz[0])
        len_reduced = traj_reduced.path_length()
        self.assertEqual(len_initial_segment, len_reduced)

    def test_transform(self):
        traj_transformed = copy.deepcopy(ex_kitti_traj)
        t = lie.random_se3()
        traj_transformed.transform(t)
        # traj_transformed.transform(lie.se3_inverse(t))
        self.assertAlmostEqual(traj_transformed.path_length(),
                               ex_kitti_traj.path_length())

    def test_scale(self):
        traj_scaled = copy.deepcopy(ex_kitti_traj)
        s = 5.234
        traj_scaled.scale(s)
        len_initial = ex_kitti_traj.path_length()
        len_scaled = traj_scaled.path_length()
        self.assertAlmostEqual(len_initial * s, len_scaled)

    def test_check(self):
        self.assertTrue(ex_kitti_traj.check()[0])
        self.assertFalse(ex_kitti_traj_wrong.check()[0])

    def test_get_infos(self):
        ex_kitti_traj.get_infos()

    def test_get_statistics(self):
        ex_kitti_traj.get_statistics()


class TestPoseTrajectory3D(unittest.TestCase):
    def test_equals(self):
        ex_tum_traj_copy = copy.deepcopy(ex_tum_traj)
        self.assertTrue(ex_tum_traj == ex_tum_traj_copy)
        self.assertFalse(ex_tum_traj == ex_tum_traj_wrong_stamps)
        self.assertFalse(ex_tum_traj == ex_tum_traj_wrong_quat)
        self.assertTrue(ex_tum_traj != ex_tum_traj_wrong_stamps)
        self.assertTrue(ex_tum_traj != ex_tum_traj_wrong_quat)
        self.assertFalse(ex_tum_traj != ex_tum_traj_copy)

    def test_reduce_to_ids(self):
        traj_reduced = copy.deepcopy(ex_tum_traj)
        traj_reduced.reduce_to_ids([0, 2])
        self.assertEqual(traj_reduced.num_poses, 2)
        self.assertEqual(len(traj_reduced.timestamps), 2)

    def test_check(self):
        self.assertTrue(ex_tum_traj.check()[0])
        self.assertFalse(ex_tum_traj_wrong_quat.check()[0])
        self.assertFalse(ex_tum_traj_wrong_stamps.check()[0])

    def test_get_infos(self):
        ex_tum_traj.get_infos()

    def test_get_statistics(self):
        ex_tum_traj.get_statistics()


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python

from evo import main_evaluation as me

if __name__ == '__main__':
    list_of_datasets = [\
                        'MH_01_easy',
                        'MH_02_easy',
                        'MH_03_medium',
                        'mh_04_difficult', # Diff number of left/right imgs...
                        'MH_05_difficult',
                        'V1_01_easy',
                        'V1_02_medium',
                        'V1_03_difficult',
                        'V2_01_easy',
                        'V2_02_medium',
                        'v2_03_difficult' # Diff number of left/right imgs...
                       ]

    list_of_pipelines = ["S", "SP", "SPR"]

    me.aggregate_ape_results(list_of_datasets, list_of_pipelines)



# -*- coding: UTF8 -*-
"""
Low- and high-level read/write functions for different file formats.
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import csv
import io
import json
import logging
import os
import zipfile

import numpy as np

import evo.core.lie_algebra as lie
import evo.core.transformations as tr
from evo.core import result
from evo.core.trajectory import PosePath3D, PoseTrajectory3D
from evo.tools import user

logger = logging.getLogger(__name__)


class FileInterfaceException(Exception):
    pass


def csv_read_matrix(file_path, delim=',', comment_str="#"):
    """
    directly parse a csv-like file into a matrix
    :param file_path: path of csv file (or file handle)
    :param delim: delimiter character
    :param comment_str: string indicating a comment line to ignore
    :return: 2D list with raw data (string)
    """
    if hasattr(file_path, 'read'):  # if file handle
        generator = (line for line in file_path
                     if not line.startswith(comment_str))
        reader = csv.reader(generator, delimiter=delim)
        mat = [row for row in reader]
    else:
        if not os.path.isfile(file_path):
            raise FileInterfaceException("csv file " + str(file_path) +
                                         " does not exist")
        with open(file_path) as f:
            generator = (line for line in f
                         if not line.startswith(comment_str))
            reader = csv.reader(generator, delimiter=delim)
            mat = [row for row in reader]
    return mat


def read_tum_trajectory_file(file_path):
    """
    parses trajectory file in TUM format (timestamp tx ty tz qx qy qz qw)
    :param file_path: the trajectory file path (or file handle)
    :return: trajectory.PoseTrajectory3D object
    """
    raw_mat = csv_read_matrix(file_path, delim=" ", comment_str="#")
    error_msg = ("TUM trajectory files must have 8 entries per row ",
                 "and no trailing delimiter at the end of the rows (space)")
    if len(raw_mat) > 0 and len(raw_mat[0]) != 8:
        raise FileInterfaceException(error_msg)
    try:
        mat = np.array(raw_mat).astype(float)
    except ValueError:
        raise FileInterfaceException(error_msg)
    stamps = mat[:, 0]  # n x 1
    xyz = mat[:, 1:4]  # n x 3
    quat = mat[:, 4:]  # n x 4
    quat = np.roll(quat, 1, axis=1)  # shift 1 column -> w in front column
    if not hasattr(file_path, 'read'):  # if not file handle
        logger.debug("Loaded {} stamps and poses from: {}".format(
            len(stamps), file_path))
    return PoseTrajectory3D(xyz, quat, stamps)


def write_tum_trajectory_file(file_path, traj, confirm_overwrite=False):
    """
    :param file_path: desired text file for trajectory (string or handle)
    :param traj: trajectory.PoseTrajectory3D
    :param confirm_overwrite: whether to require user interaction
           to overwrite existing files
    """
    if isinstance(file_path, str) and confirm_overwrite:
        if not user.check_and_confirm_overwrite(file_path):
            return
    if not isinstance(traj, PoseTrajectory3D):
        raise FileInterfaceException(
            "trajectory must be a PoseTrajectory3D object")
    stamps = traj.timestamps
    xyz = traj.positions_xyz
    # shift -1 column -> w in back column
    quat = np.roll(traj.orientations_quat_wxyz, -1, axis=1)
    mat = np.column_stack((stamps, xyz, quat))
    np.savetxt(file_path, mat, delimiter=" ")
    if isinstance(file_path, str):
        logger.info("Trajectory saved to: " + file_path)


def read_kitti_poses_file(file_path):
    """
    parses pose file in KITTI format (first 3 rows of SE(3) matrix per line)
    :param file_path: the trajectory file path (or file handle)
    :return: trajectory.PosePath3D
    """
    raw_mat = csv_read_matrix(file_path, delim=" ", comment_str="#")
    error_msg = ("KITTI pose files must have 12 entries per row ",
                 "and no trailing delimiter at the end of the rows (space)")
    if len(raw_mat) > 0 and len(raw_mat[0]) != 12:
        raise FileInterfaceException(error_msg)
    try:
        mat = np.array(raw_mat).astype(float)
    except ValueError:
        raise FileInterfaceException(error_msg)
    # yapf: disable
    poses = [np.array([[r[0], r[1], r[2], r[3]],
                       [r[4], r[5], r[6], r[7]],
                       [r[8], r[9], r[10], r[11]],
                       [0, 0, 0, 1]]) for r in mat]
    # yapf: enable
    if not hasattr(file_path, 'read'):  # if not file handle
        logger.debug("Loaded {} poses from: {}".format(len(poses), file_path))
    return PosePath3D(poses_se3=poses)


def write_kitti_poses_file(file_path, traj, confirm_overwrite=False):
    """
    :param file_path: desired text file for trajectory (string or handle)
    :param traj: trajectory.PosePath3D or trajectory.PoseTrajectory3D
    :param confirm_overwrite: whether to require user interaction
           to overwrite existing files
    """
    if isinstance(file_path, str) and confirm_overwrite:
        if not user.check_and_confirm_overwrite(file_path):
            return
    # first 3 rows  of SE(3) matrix flattened
    poses_flat = [p.flatten()[:-4] for p in traj.poses_se3]
    np.savetxt(file_path, poses_flat, delimiter=' ')
    if isinstance(file_path, str):
        logger.info("Poses saved to: " + file_path)


def read_euroc_csv_trajectory(file_path):
    """
    parses ground truth trajectory from EuRoC MAV state estimate .csv
    :param file_path: <sequence>/mav0/state_groundtruth_estimate0/data.csv
    :return: trajectory.PoseTrajectory3D object
    """
    raw_mat = csv_read_matrix(file_path, delim=",", comment_str="#")
    error_msg = ("EuRoC MAV state ground truth must have 17 entries per row ",
                 "and no trailing delimiter at the end of the rows (comma)")
    if len(raw_mat) > 0 and len(raw_mat[0]) != 17:
        raise FileInterfaceException(error_msg)
    try:
        mat = np.array(raw_mat).astype(float)
    except ValueError:
        raise FileInterfaceException(error_msg)
    stamps = np.divide(mat[:, 0], 1e9)  # n x 1  -  nanoseconds to seconds
    xyz = mat[:, 1:4]  # n x 3
    quat = mat[:, 4:8]  # n x 4
    logger.debug("Loaded {} stamps and poses from: {}".format(
        len(stamps), file_path))
    return PoseTrajectory3D(xyz, quat, stamps)


def read_swe_csv_trajectory(file_path):
    """
    parses trajectory file in SWE format (timestamp[ns], x, y, z, qx, qy, qz, qw, vx, vy, vz, bgx, bgy, bgz, bax, bay, baz)
    :param file_path: the trajectory file path (or file handle)
    :return: trajectory.PoseTrajectory3D object
    """
    mat = np.array(csv_read_matrix(
        file_path, delim=",", comment_str="t")).astype(float)
    if mat.shape[1] != 17:
        raise FileInterfaceException(
            "SWE trajectory files must have 8 entries per row")
    stamps = np.divide(mat[:, 0], 1e9)  # n x 1  -  nanoseconds to seconds
    xyz = mat[:, 1:4]  # n x 3
    quat = mat[:, 4:8]  # n x 4
    quat = np.roll(quat, 1, axis=1)  # shift 1 column -> w in front column
    if not hasattr(file_path, 'read'):  # if not file handle
        logger.debug("Loaded {} stamps and poses from: {}".format(
            len(stamps), file_path))
    return PoseTrajectory3D(xyz, quat, stamps)


def read_bag_trajectory(bag_handle, topic):
    """
    :param bag_handle: opened bag handle, from rosbag.Bag(...)
    :param topic: geometry_msgs/PoseStamped or nav_msgs/Odometry topic
    :return: trajectory.PoseTrajectory3D
    """
    if not bag_handle.get_message_count(topic) > 0:
        raise FileInterfaceException("no messages for topic '" + topic +
                                     "' in bag")
    msg_type = bag_handle.get_type_and_topic_info().topics[topic].msg_type
    if msg_type not in {"geometry_msgs/PoseStamped", "nav_msgs/Odometry"}:
        raise FileInterfaceException(
            "unsupported message type: {}".format(msg_type))
    stamps, xyz, quat = [], [], []
    for topic, msg, t in bag_handle.read_messages(topic):
        stamps.append(t.secs + (t.nsecs * 1e-9))
        # Make nav_msgs/Odometry behave like geometry_msgs/PoseStamped.
        while not hasattr(msg.pose, 'position') and not hasattr(
                msg.pose, 'orientation'):
            msg = msg.pose
        xyz.append(
            [msg.pose.position.x, msg.pose.position.y, msg.pose.position.z])
        quat.append([
            msg.pose.orientation.x, msg.pose.orientation.y,
            msg.pose.orientation.z, msg.pose.orientation.w
        ])
    quat = np.roll(quat, 1, axis=1)  # shift 1 column -> w in front column
    logger.debug("Loaded {} {} messages of topic: {}".format(
        len(stamps), msg_type, topic))
    generator = bag_handle.read_messages(topic)
    _, first_msg, _ = generator.next()
    frame_id = first_msg.header.frame_id
    return PoseTrajectory3D(xyz, quat, stamps, meta={"frame_id": frame_id})


def write_bag_trajectory(bag_handle, traj, topic_name, frame_id=""):
    """
    :param bag_handle: opened bag handle, from rosbag.Bag(...)
    :param traj: trajectory.PoseTrajectory3D
    :param topic_name: the desired topic name for the trajectory
    :param frame_id: optional ROS frame_id
    """
    import rospy
    from geometry_msgs.msg import PoseStamped
    if not isinstance(traj, PoseTrajectory3D):
        raise FileInterfaceException(
            "trajectory must be a PoseTrajectory3D object")
    for stamp, xyz, quat in zip(traj.timestamps, traj.positions_xyz,
                                traj.orientations_quat_wxyz):
        p = PoseStamped()
        p.header.stamp = rospy.Time.from_sec(stamp)
        p.header.frame_id = frame_id
        p.pose.position.x = xyz[0]
        p.pose.position.y = xyz[1]
        p.pose.position.z = xyz[2]
        p.pose.orientation.w = quat[0]
        p.pose.orientation.x = quat[1]
        p.pose.orientation.y = quat[2]
        p.pose.orientation.z = quat[3]
        bag_handle.write(topic_name, p, t=p.header.stamp)
    logger.info("Saved geometry_msgs/PoseStamped topic: " + topic_name)


def save_res_file(zip_path, result_obj, confirm_overwrite=False):
    """
    save results of a pose error metric (pe_metric) to a zip file
    :param zip_path: path to zip file
    :param result_obj: evo.core.result.Result instance
    :param confirm_overwrite: whether to require user interaction
           to overwrite existing files
    """
    from tempfile import TemporaryFile
    logger.debug("Saving results to " + zip_path + "...")
    if confirm_overwrite and not user.check_and_confirm_overwrite(zip_path):
        return
    with zipfile.ZipFile(zip_path, 'w') as archive:
        archive.writestr("info.json", json.dumps(result_obj.info))
        archive.writestr("stats.json", json.dumps(result_obj.stats))
        for name, array in result_obj.np_arrays.items():
            tmp_file = TemporaryFile()
            np.save(tmp_file, array)
            tmp_file.seek(0)
            archive.writestr("{}.npz".format(name), tmp_file.read())
            tmp_file.close()
        for name, traj in result_obj.trajectories.items():
            tmp_file = TemporaryFile()
            if type(traj) is PosePath3D:
                fmt_suffix = ".kitti"
                write_kitti_poses_file(tmp_file, traj)
            elif type(traj) is PoseTrajectory3D:
                fmt_suffix = ".tum"
                write_tum_trajectory_file(tmp_file, traj)
            else:
                raise FileInterfaceException(
                    "unknown format of trajectory {}".format(name))
            tmp_file.seek(0)
            archive.writestr("{}{}".format(name, fmt_suffix), tmp_file.read())
            tmp_file.close()


def load_res_file(zip_path, load_trajectories=False):
    """
    load contents of a result .zip file saved with save_res_file(...)
    :param zip_path: path to zip file
    :param load_trajectories: set to True to load also the (backup) trajectories
    :return: evo.core.result.Result instance
    """
    logger.debug("Loading result from {} ...".format(zip_path))
    result_obj = result.Result()
    with zipfile.ZipFile(zip_path, mode='r') as archive:
        file_list = archive.namelist()
        if not {"error_array.npz", "info.json", "stats.json"
                } <= set(file_list):
            logger.warning("{} not suitable for evo_res".format(zip_path))
        npz_files = [n for n in archive.namelist() if n.endswith(".npz")]
        for filename in npz_files:
            with io.BytesIO(archive.read(filename)) as f:
                array = np.load(f)
                name = os.path.splitext(os.path.basename(filename))[0]
                result_obj.add_np_array(name, array)
        if load_trajectories:
            tum_files = [n for n in archive.namelist() if n.endswith(".tum")]
            for filename in tum_files:
                with io.TextIOWrapper(archive.open(filename, mode='r')) as f:
                    traj = read_tum_trajectory_file(f)
                    name = os.path.splitext(os.path.basename(filename))[0]
                    result_obj.add_trajectory(name, traj)
            kitti_files = [
                n for n in archive.namelist() if n.endswith(".kitti")
            ]
            for filename in kitti_files:
                with io.TextIOWrapper(archive.open(filename, mode='r')) as f:
                    traj = read_kitti_poses_file(f)
                    name = os.path.splitext(os.path.basename(filename))[0]
                    result_obj.add_trajectory(name, traj)
        result_obj.info = json.loads(archive.read("info.json").decode("utf-8"))
        result_obj.stats = json.loads(
            archive.read("stats.json").decode("utf-8"))
    return result_obj


def load_transform_json(json_path):
    """
    load a transformation stored in xyz + quaternion format in a .json file
    :param json_path: path to the .json file
    :return: t (SE(3) matrix)
    """
    with open(json_path, 'r') as tf_file:
        data = json.load(tf_file)
        keys = ("x", "y", "z", "qx", "qy", "qz", "qw")
        if not all(key in data for key in keys):
            raise FileInterfaceException(
                "invalid transform file - expected keys " + str(keys))
        xyz = np.array([data["x"], data["y"], data["z"]])
        quat = np.array([data["qw"], data["qx"], data["qy"], data["qz"]])
        t = lie.se3(lie.so3_from_se3(tr.quaternion_matrix(quat)), xyz)
        return t



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main package executable
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import os
from evo.tools import settings

DESC = '''
Python package for the evaluation of odometry and SLAM
(c) michael.grupp@tum.de - license: run 'evo pkg --license'

Supported trajectory formats:
* TUM trajectory files
* KITTI pose files
* ROS bagfile with geometry_msgs/PoseStamped or nav_msgs/Odometry topics
* EuRoC MAV dataset groundtruth files

The following executables are available:

Metrics:
   evo_ape - absolute pose error
   evo_rpe - relative pose error

Tools:
   evo_traj - tool for analyzing, plotting or exporting multiple trajectories
   evo_res - tool for processing multiple result files from the metrics
   evo_ipython - IPython shell with pre-loaded evo modules
   evo_fig - (experimental) tool for re-opening serialized plots
   evo_config - tool for global settings and config file manipulation
'''


def main():
    import sys
    import argparse
    import argcomplete
    main_parser = argparse.ArgumentParser()
    shared_parser = argparse.ArgumentParser(add_help=False)
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True
    pkg_parser = sub_parsers.add_parser(
        "pkg", description="show infos of the package",
        parents=[shared_parser])
    pkg_parser.add_argument("--info", help="show the package description",
                            action="store_true")
    pkg_parser.add_argument("--version", help="print the package version",
                            action="store_true")
    pkg_parser.add_argument("--pyversion", help="print the Python version",
                            action="store_true")
    pkg_parser.add_argument("--license", help="print the package license",
                            action="store_true")
    pkg_parser.add_argument("--location", help="print the package path",
                            action="store_true")
    pkg_parser.add_argument("--logfile", help="print the logfile path",
                            action="store_true")
    pkg_parser.add_argument("--open_log", help="open the package logfile",
                            action="store_true")
    pkg_parser.add_argument("--clear_log", help="clear package logfile",
                            action="store_true")
    cat_parser = sub_parsers.add_parser(
        "cat_log", description="pipe stdin to evo logfile"
        " or print logfile to stdout (if no stdin)", parents=[shared_parser])
    cat_parser.add_argument("-l", "--loglevel", help="loglevel of the message",
                            default="info",
                            choices=["error", "warning", "info", "debug"])
    cat_parser.add_argument("-m", "--message",
                            help="explicit message instead of pipe")
    cat_parser.add_argument("-s", "--source",
                            help="source name to use for the log message")
    cat_parser.add_argument("--clear_log", help="clear logfile before exiting",
                            action="store_true")
    argcomplete.autocomplete(main_parser)
    if len(sys.argv[1:]) == 0:
        sys.argv.extend(["pkg", "--info"])  # cheap trick because YOLO
    args = main_parser.parse_args()
    line_end = "\n" if sys.stdout.isatty() else ""

    if args.subcommand == "pkg":
        if not len(sys.argv) > 2:
            pkg_parser.print_help()
            sys.exit(1)
        if args.license:
            print(open(os.path.join(settings.PACKAGE_BASE_PATH,
                                    "LICENSE")).read())
        if args.info:
            main_parser.print_usage()
            print(DESC)
        if args.version:
            here = os.path.dirname(os.path.abspath(__file__))
            version_path = os.path.join(here, "version")
            print(" ".join(open(version_path).read().splitlines()),
                  end=line_end)
        if args.pyversion:
            import platform as pf
            print(pf.python_version(), end=line_end)
        if args.location:
            print(settings.PACKAGE_BASE_PATH, end=line_end)
        if args.logfile or args.open_log:
            if not os.path.exists(settings.DEFAULT_LOGFILE_PATH):
                print("no logfile found - run: evo_config set logfile_enabled",
                      end=line_end)
                sys.exit(1)
            print(settings.DEFAULT_LOGFILE_PATH, end=line_end)
            if args.open_log:
                import webbrowser
                webbrowser.open(settings.DEFAULT_LOGFILE_PATH)
        if args.clear_log:
            from evo.tools import user
            if user.confirm("clear logfile? (y/n)"):
                open(settings.DEFAULT_LOGFILE_PATH, mode='w')

    elif args.subcommand == "cat_log":
        if os.name == "nt":
            print("cat_log feature not available on Windows")
            sys.exit(1)
        if not args.message and sys.stdin.isatty():
            if not os.path.exists(settings.DEFAULT_LOGFILE_PATH):
                print("no logfile found - run: evo_config set logfile_enabled",
                      end=line_end)
            else:
                print(open(settings.DEFAULT_LOGFILE_PATH).read(), end="")
        elif not settings.SETTINGS.logfile_enabled:
            print("logfile disabled", end=line_end)
            sys.exit(1)
        else:
            import logging
            logger = logging.getLogger(__name__)
            from evo.tools import log
            file_fmt = log.DEFAULT_LONG_FMT
            if args.source:
                file_fmt = file_fmt.replace(
                    "%(module)s.%(funcName)s():%(lineno)s", args.source)
            log.configure_logging(silent=True, file_fmt=file_fmt)
            if not args.message:
                msg = sys.stdin.read()
            else:
                msg = args.message
            getattr(logger, args.loglevel)(msg)
        if args.clear_log:
            open(settings.DEFAULT_LOGFILE_PATH, mode='w')


if __name__ == '__main__':
    main()



"""
Common functions for evo_ape and evo_rpe, internal only.
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging

logger = logging.getLogger(__name__)

SEP = "-" * 80  # separator line


def load_trajectories(args):
    from evo.core import sync
    from evo.tools import file_interface

    if args.subcommand == "tum":
        traj_ref = file_interface.read_tum_trajectory_file(args.ref_file)
        traj_est = file_interface.read_tum_trajectory_file(args.est_file)
        ref_name, est_name = args.ref_file, args.est_file
    elif args.subcommand == "kitti":
        traj_ref = file_interface.read_kitti_poses_file(args.ref_file)
        traj_est = file_interface.read_kitti_poses_file(args.est_file)
        ref_name, est_name = args.ref_file, args.est_file
    elif args.subcommand == "euroc":
        args.align = True
        logger.info("Forcing trajectory alignment implicitly "
                    "(EuRoC ground truth is in IMU frame).")
        logger.debug(SEP)
        traj_ref = file_interface.read_euroc_csv_trajectory(args.state_gt_csv)
        traj_est = file_interface.read_tum_trajectory_file(args.est_file)
        ref_name, est_name = args.state_gt_csv, args.est_file
    elif args.subcommand == "bag":
        import rosbag
        logger.debug("Opening bag file " + args.bag)
        bag = rosbag.Bag(args.bag, 'r')
        try:
            traj_ref = file_interface.read_bag_trajectory(bag, args.ref_topic)
            traj_est = file_interface.read_bag_trajectory(bag, args.est_topic)
            ref_name, est_name = args.ref_topic, args.est_topic
        finally:
            bag.close()
    else:
        raise KeyError("unknown sub-command: {}".format(args.subcommand))

    if args.subcommand != "kitti":
        logger.debug("Synchronizing trajectories...")
        traj_ref, traj_est = sync.associate_trajectories(
            traj_ref, traj_est, args.t_max_diff, args.t_offset,
            first_name=ref_name, snd_name=est_name)

    return traj_ref, traj_est, ref_name, est_name


def get_pose_relation(args):
    from evo.core.metrics import PoseRelation
    pose_relation = None
    if args.pose_relation == "full":
        pose_relation = PoseRelation.full_transformation
    elif args.pose_relation == "rot_part":
        pose_relation = PoseRelation.rotation_part
    elif args.pose_relation == "trans_part":
        pose_relation = PoseRelation.translation_part
    elif args.pose_relation == "angle_deg":
        pose_relation = PoseRelation.rotation_angle_deg
    elif args.pose_relation == "angle_rad":
        pose_relation = PoseRelation.rotation_angle_rad
    return pose_relation


def get_delta_unit(args):
    from evo.core.metrics import Unit
    delta_unit = None
    if args.delta_unit == "f":
        delta_unit = Unit.frames
    elif args.delta_unit == "d":
        delta_unit = Unit.degrees
    elif args.delta_unit == "r":
        delta_unit = Unit.radians
    elif args.delta_unit == "m":
        delta_unit = Unit.meters
    return delta_unit


def plot(args, result, traj_ref, traj_est):
    from evo.tools import plot
    from evo.tools.settings import SETTINGS

    import matplotlib.pyplot as plt
    import numpy as np

    logger.debug(SEP)
    logger.debug("Plotting results... ")
    plot_mode = plot.PlotMode(args.plot_mode)

    # Plot the raw metric values.
    fig1 = plt.figure(figsize=SETTINGS.plot_figsize)
    if "seconds_from_start" in result.np_arrays:
        seconds_from_start = result.np_arrays["seconds_from_start"]
    else:
        seconds_from_start = None

    plot.error_array(fig1, result.np_arrays["error_array"],
                     x_array=seconds_from_start, statistics=result.stats,
                     name=result.info["label"], title=result.info["title"],
                     xlabel="$t$ (s)" if seconds_from_start else "index")

    # Plot the values color-mapped onto the trajectory.
    fig2 = plt.figure(figsize=SETTINGS.plot_figsize)
    ax = plot.prepare_axis(fig2, plot_mode)
    plot.traj(ax, plot_mode, traj_ref, '--', 'black', 'reference',
              alpha=0.0 if SETTINGS.plot_hideref else 0.5)

    if args.plot_colormap_min is None:
        args.plot_colormap_min = result.stats["min"]
    if args.plot_colormap_max is None:
        args.plot_colormap_max = result.stats["max"]
    if args.plot_colormap_max_percentile is not None:
        args.plot_colormap_max = np.percentile(
            result.np_arrays["error_array"], args.plot_colormap_max_percentile)

    plot.traj_colormap(ax, traj_est, result.np_arrays["error_array"],
                       plot_mode, min_map=args.plot_colormap_min,
                       max_map=args.plot_colormap_max,
                       title="Error mapped onto trajectory")
    fig2.axes.append(ax)

    plot_collection = plot.PlotCollection(result.info["title"])
    plot_collection.add_figure("raw", fig1)
    plot_collection.add_figure("map", fig2)
    if args.plot:
        plot_collection.show()
    if args.save_plot:
        plot_collection.export(args.save_plot,
                               confirm_overwrite=not args.no_warnings)
    if args.serialize_plot:
        logger.debug(SEP)
        plot_collection.serialize(args.serialize_plot,
                                  confirm_overwrite=not args.no_warnings)



# -*- coding: UTF8 -*-
"""
some functions for trajectories
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import logging

import numpy as np

import evo.core.transformations as tr
import evo.core.geometry as geometry
from evo.core import lie_algebra as lie

logger = logging.getLogger(__name__)


class TrajectoryException(Exception):
    pass


class PosePath3D(object):
    """
    just a path, no temporal information
    also: base class for real trajectory
    """

    def __init__(self, positions_xyz=None, orientations_quat_wxyz=None,
                 poses_se3=None, meta=None):
        """
        :param positions_xyz: nx3 list of x,y,z positions
        :param orientations_quat_wxyz: nx4 list of quaternions (w,x,y,z format)
        :param poses_se3: list of SE(3) poses
        :param meta: optional metadata
        """
        if (positions_xyz is None
                or orientations_quat_wxyz is None) and poses_se3 is None:
            raise TrajectoryException("must provide at least positions_xyz "
                                      "& orientations_quat_wxyz or poses_se3")
        if positions_xyz is not None:
            self._positions_xyz = np.array(positions_xyz)
        if orientations_quat_wxyz is not None:
            self._orientations_quat_wxyz = np.array(orientations_quat_wxyz)
        if poses_se3 is not None:
            self._poses_se3 = poses_se3
        self.meta = {} if meta is None else meta

    def __str__(self):
        return "{} poses, {:.3f}m path length".format(self.num_poses,
                                                      self.path_length())

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        if not self.num_poses == other.num_poses:
            return False
        equal = True
        equal &= all([
            np.allclose(p1, p2)
            for p1, p2 in zip(self.poses_se3, other.poses_se3)
        ])
        equal &= np.allclose(self.orientations_quat_wxyz,
                             other.orientations_quat_wxyz)
        equal &= np.allclose(self.positions_xyz, other.positions_xyz)
        return equal

    def __ne__(self, other):
        return not self == other

    @property
    def positions_xyz(self):
        if not hasattr(self, "_positions_xyz"):
            assert hasattr(self, "_poses_se3")
            self._positions_xyz = np.array([p[:3, 3] for p in self._poses_se3])
        return self._positions_xyz

    @property
    def orientations_quat_wxyz(self):
        if not hasattr(self, "_orientations_quat_wxyz"):
            assert hasattr(self, "_poses_se3")
            self._orientations_quat_wxyz \
                = np.array(
                    [tr.quaternion_from_matrix(p)
                     for p in self._poses_se3])
        return self._orientations_quat_wxyz

    @property
    def orientations_euler(self):
        if not hasattr(self, "_orientations_euler"):
            if hasattr(self, "_poses_se3"):
                self._orientations_euler \
                    = np.array(
                        [tr.euler_from_matrix(p, axes="sxyz")
                         for p in self._poses_se3])
            elif hasattr(self, "_orientations_quat_wxyz"):
                self._orientations_euler \
                    = np.array([tr.euler_from_quaternion(q, axes="sxyz")
                                for q in self._orientations_quat_wxyz])
        return self._orientations_euler

    @property
    def poses_se3(self):
        if not hasattr(self, "_poses_se3"):
            assert hasattr(self, "_positions_xyz")
            assert hasattr(self, "_orientations_quat_wxyz")
            self._poses_se3 \
                = xyz_quat_wxyz_to_se3_poses(self.positions_xyz,
                                             self.orientations_quat_wxyz)
        return self._poses_se3

    @property
    def num_poses(self):
        if hasattr(self, "_poses_se3"):
            return len(self._poses_se3)
        else:
            return self.positions_xyz.shape[0]

    def path_length(self, ids=None):
        """
        calculates the path length (arc-length)
        :param ids: optional start and end index as tuple (start, end)
        :return: path length in meters
        """
        if ids is not None:
            if len(ids) != 2 or not all(type(i) is int for i in ids):
                raise TrajectoryException(
                    "ids must be a tuple of positive integers")
            return float(geometry.arc_len(self.positions_xyz[ids[0]:ids[1]]))
        else:
            return float(geometry.arc_len(self.positions_xyz))

    def transform(self, t, right_mul=False):
        """
        apply a left or right multiplicative SE(3) transformation to the whole path
        :param t: a valid SE(3) matrix
        :param right_mul: whether to apply it right-multiplicative or not
        """
        if not lie.is_se3(t):
            raise TrajectoryException(
                "transformation is not a valid SE(3) matrix")
        if right_mul:
            self._poses_se3 = [np.dot(p, t) for p in self.poses_se3]
        else:
            self._poses_se3 = [np.dot(t, p) for p in self.poses_se3]
        self._positions_xyz, self._orientations_quat_wxyz \
            = se3_poses_to_xyz_quat_wxyz(self.poses_se3)

    def scale(self, s):
        """
        apply a scaling to the whole path
        :param s: scale factor
        """
        if hasattr(self, "_poses_se3"):
            self._poses_se3 = [
                lie.se3(p[:3, :3], s * p[:3, 3]) for p in self._poses_se3
            ]
        if hasattr(self, "_positions_xyz"):
            self._positions_xyz = s * self._positions_xyz

    def reduce_to_ids(self, ids):
        """
        reduce the elements to the ones specified in ids
        :param ids: list of integer indices
        """
        if hasattr(self, "_positions_xyz"):
            self._positions_xyz = self._positions_xyz[ids]
        if hasattr(self, "_orientations_quat_wxyz"):
            self._orientations_quat_wxyz = self._orientations_quat_wxyz[ids]
        if hasattr(self, "_poses_se3"):
            self._poses_se3 = [self._poses_se3[idx] for idx in ids]

    def check(self):
        """
        checks if the data is valid
        :return: True/False, dictionary with some detailed infos
        """
        same_len = self.positions_xyz.shape[0] \
            == self.orientations_quat_wxyz.shape[0] \
            == len(self.poses_se3)
        se3_valid = all([lie.is_se3(p) for p in self.poses_se3])
        norms = np.linalg.norm(self.orientations_quat_wxyz, axis=1)
        quat_normed = np.allclose(norms, np.ones(norms.shape))
        valid = same_len and se3_valid and quat_normed
        details = {
            "array shapes": "ok"
            if same_len else "invalid (lists must have same length)",
            "SE(3) conform": "yes"
            if se3_valid else "no (poses are not valid SE(3) matrices)",
            "quaternions": "ok"
            if quat_normed else "invalid (must be unit quaternions)"
        }
        return valid, details

    def get_infos(self):
        """
        :return: dictionary with some infos about the path
        """
        return {
            "nr. of poses": self.num_poses,
            "path length (m)": self.path_length(),
            "pos_start (m)": self.positions_xyz[0],
            "pos_end (m)": self.positions_xyz[-1]
        }

    def get_statistics(self):
        return {}  # no idea yet


class PoseTrajectory3D(PosePath3D, object):
    """
    a PosePath with temporal information
    """

    def __init__(self, positions_xyz=None, orientations_quat_wxyz=None,
                 timestamps=None, poses_se3=None, meta=None):
        """
        :param timestamps: optional nx1 list of timestamps
        """
        super(PoseTrajectory3D, self).__init__(
            positions_xyz, orientations_quat_wxyz, poses_se3, meta)
        # this is a bit ugly...
        if timestamps is None:
            raise TrajectoryException("no timestamps provided")
        self.timestamps = np.array(timestamps)

    def __str__(self):
        s = super(PoseTrajectory3D, self).__str__()
        return s + ", {:.3f}s duration".format(self.timestamps[-1] -
                                               self.timestamps[0])

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        if not self.num_poses == other.num_poses:
            return False
        equal = super(PoseTrajectory3D, self).__eq__(other)
        equal &= np.allclose(self.timestamps, other.timestamps)
        return equal

    def __ne__(self, other):
        return not self == other

    def reduce_to_ids(self, ids):
        super(PoseTrajectory3D, self).reduce_to_ids(ids)
        self.timestamps = self.timestamps[ids]

    def check(self):
        valid, details = super(PoseTrajectory3D, self).check()
        len_stamps_valid = (len(self.timestamps) == len(self.positions_xyz))
        valid &= len_stamps_valid
        details["nr. of stamps"] = "ok" if len_stamps_valid else "wrong"
        stamps_ascending = np.alltrue(
            np.sort(self.timestamps) == self.timestamps)
        stamps_ascending &= np.unique(self.timestamps).size == len(
            self.timestamps)
        valid &= stamps_ascending
        if stamps_ascending:
            details["timestamps"] = "ok"
        else:
            details["timestamps"] = "wrong, not ascending or duplicates"
        return valid, details

    def get_infos(self):
        """
        :return: dictionary with some infos about the trajectory
        """
        infos = super(PoseTrajectory3D, self).get_infos()
        infos["duration (s)"] = self.timestamps[-1] - self.timestamps[0]
        infos["t_start (s)"] = self.timestamps[0]
        infos["t_end (s)"] = self.timestamps[-1]
        return infos

    def get_statistics(self):
        """
        :return: dictionary with some statistics of the trajectory
        """
        stats = super(PoseTrajectory3D, self).get_statistics()
        speeds = [
            calc_speed(self.positions_xyz[i], self.positions_xyz[i + 1],
                       self.timestamps[i], self.timestamps[i + 1])
            for i in range(len(self.positions_xyz) - 1)
        ]
        vmax = max(speeds)
        vmin = min(speeds)
        vmean = np.mean(speeds)
        stats.update({
            "v_max (m/s)": vmax,
            "v_min (m/s)": vmin,
            "v_avg (m/s)": vmean,
            "v_max (km/h)": vmax * 3.6,
            "v_min (km/h)": vmin * 3.6,
            "v_avg (km/h)": vmean * 3.6
        })
        return stats


class Trajectory(PoseTrajectory3D):
    pass  # TODO compat


def calc_speed(xyz_1, xyz_2, t_1, t_2):
    """
    :param xyz_1: position at timestamp 1
    :param xyz_2: position at timestamp 2
    :param t_1: timestamp 1
    :param t_2: timestamp 2
    :return: speed in m/s
    """
    if (t_2 - t_1) <= 0:
        raise TrajectoryException("bad timestamps: " + str(t_1) + " & " +
                                  str(t_2))
    return np.linalg.norm(xyz_2 - xyz_1) / (t_2 - t_1)


def calc_angular_speed(p_1, p_2, t_1, t_2, degrees=False):
    """
    :param p_1: pose at timestamp 1
    :param p_2: pose at timestamp 2
    :param t_1: timestamp 1
    :param t_2: timestamp 2
    :param degrees: set to True to return deg/s
    :return: speed in rad/s
    """
    if (t_2 - t_1) <= 0:
        raise TrajectoryException("bad timestamps: " + str(t_1) + " & " +
                                  str(t_2))
    if degrees:
        angle_1 = lie.so3_log(p_1[:3, :3]) * 180 / np.pi
        angle_2 = lie.so3_log(p_2[:3, :3]) * 180 / np.pi
    else:
        angle_1 = lie.so3_log(p_1[:3, :3])
        angle_2 = lie.so3_log(p_2[:3, :3])
    return (angle_2 - angle_1) / (t_2 - t_1)


def xyz_quat_wxyz_to_se3_poses(xyz, quat):
    poses = [
        lie.se3(lie.so3_from_se3(tr.quaternion_matrix(quat)), xyz)
        for quat, xyz in zip(quat, xyz)
    ]
    return poses


def se3_poses_to_xyz_quat_wxyz(poses):
    xyz = np.array([pose[:3, 3] for pose in poses])
    quat_wxyz = np.array([tr.quaternion_from_matrix(pose) for pose in poses])
    return xyz, quat_wxyz


def align_trajectory(traj, traj_ref, correct_scale=False, correct_only_scale=False,
                     return_parameters=False, n=-1,
                     discard_n_start_poses=0, discard_n_end_poses=0):
    """
    align a trajectory to a reference using Umeyama alignment
    :param traj: the trajectory to align
    :param traj_ref: reference trajectory
    :param correct_scale: set to True to adjust also the scale
    :param correct_only_scale: set to True to correct the scale, but not the pose
    :param n: the number of poses to use, counted from the start (default: all)
    :param return_parameters: also return result parameters of Umeyama's method
    :param discard_n_start_poses: the number of poses to skip from the start (default: none)
    :param discard_n_end_poses: the number of poses to discard counted from the end (default: none)
    :return: the aligned trajectory
    If return_parameters is set, the rotation matrix, translation vector and
    scaling parameter of Umeyama's method are also returned.

    """
    # otherwise np arrays will be references and mess up stuff
    traj_aligned = copy.deepcopy(traj)
    with_scale = correct_scale or correct_only_scale
    if correct_only_scale:
        logger.debug("Correcting scale...")
    else:
        logger.debug("Aligning using Umeyama's method..." +
                     (" (with scale correction)" if with_scale else ""))
        end_pose_idx = len(traj_aligned.positions_xyz)
        if n != -1:
            if discard_n_start_poses != 0 or discard_n_end_poses != 0:
                print("WARNING: you asked for %d poses, as well as discarding %d start poses and %d end poses.",
                 n, discard_n_start_poses, discard_n_end_poses)
                print("Selecting option given the smallest number of poses:")
            if end_pose_idx - discard_n_end_poses < n + discard_n_start_poses:
                print("We are requiring more poses n (%d) than available, using instead %d starting from pose %d", n,
                 end_pose_idx - discard_n_end_poses - discard_n_start_poses, discard_n_start_poses)
                end_pose_idx = end_pose_idx - discard_n_end_poses
            else:
                end_pose_idx = n + discard_n_start_poses
        else:
            end_pose_idx -= discard_n_end_poses
    assert(discard_n_start_poses < end_pose_idx) # Otherwise there will be no poses to align.
    assert(end_pose_idx <= len(traj_aligned.positions_xyz)) # Otherwise we will be requiring more poses than there are.
    r_a, t_a, s = geometry.umeyama_alignment(
        traj_aligned.positions_xyz[discard_n_start_poses:end_pose_idx, :].T,
        traj_ref.positions_xyz[discard_n_start_poses:end_pose_idx, :].T,
        with_scale)

    if not correct_only_scale:
        logger.debug("Rotation of alignment:\n{}"
                     "\nTranslation of alignment:\n{}".format(r_a, t_a))
    logger.debug("Scale correction: {}".format(s))

    if correct_only_scale:
        traj_aligned.scale(s)
    elif correct_scale:
        traj_aligned.scale(s)
        traj_aligned.transform(lie.se3(r_a, t_a))
    else:
        traj_aligned.transform(lie.se3(r_a, t_a))

    if return_parameters:
        return traj_aligned, r_a, t_a, s
    else:
        return traj_aligned


def merge(trajectories):
    """
    Merges multiple trajectories into a single, timestamp-sorted one.
    :param trajectories: list of PoseTrajectory3D objects
    :return: merged PoseTrajectory3D
    """
    merged_stamps = np.concatenate([t.timestamps for t in trajectories])
    merged_xyz = np.concatenate([t.positions_xyz for t in trajectories])
    merged_quat = np.concatenate(
        [t.orientations_quat_wxyz for t in trajectories])
    order = merged_stamps.argsort()
    merged_stamps = merged_stamps[order]
    merged_xyz = merged_xyz[order]
    merged_quat = merged_quat[order]
    return PoseTrajectory3D(merged_xyz, merged_quat, merged_stamps)



"""
default package settings definition
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import pkgutil


def get_default_plot_backend():
    backends = {"PyQt5": "Qt5Agg", "PyQt4": "Qt4Agg"}
    for pkg in backends:
        if pkgutil.find_loader(pkg) is not None:
            return backends[pkg]
    return "TkAgg"


# default settings with documentation
# yapf: disable
DEFAULT_SETTINGS_DICT_DOC = {
    "plot_xyz_realistic": (
        True,
        "Equal axes ratio in 'xyz' plot mode for realistic trajectory plots."
    ),
    "plot_backend": (
        get_default_plot_backend(),
        "matplotlib backend - default: 'Qt{4, 5}Agg' (if PyQt is installed) or 'TkAgg'."
    ),
    "plot_hideref": (
        False,
        "Hide the reference trajectory in trajectory plots."
    ),
    "plot_linewidth": (
        1.5,
        "Line width value supported by matplotlib."
    ),
    "plot_usetex": (
        False,
        "Use the LaTeX renderer configured in plot_texsystem for plots.",
    ),
    "plot_texsystem": (
        "pdflatex",
        "'xelatex', 'lualatex' or 'pdflatex', see: https://matplotlib.org/users/pgf.html",
    ),
    "plot_fontfamily": (
        "sans-serif",
        "Font family string supported by matplotlib."
    ),
    "plot_fontscale": (
        1.0,
        "Font scale value, see: https://seaborn.pydata.org/generated/seaborn.set.html"
    ),
    "plot_split": (
        False,
        "Show / save each figure separately instead of a collection."
    ),
    "plot_figsize": (
        [6, 6],
        "The default size of one (sub)plot figure (width, height)."
    ),
    "plot_trajectory_cmap": (
        "jet",
        "matplotlib color map used for mapping values on a trajectory.",
    ),
    "plot_multi_cmap": (
        "none",
        "Color map for coloring plots from multiple data sources.\n"
        + "'none' will use the default color palette, see plot_seaborn_palette."
    ),
    "plot_invert_xaxis": (
        False,
        "Invert the x-axis of plots."
    ),
    "plot_invert_yaxis": (
        False,
        "Invert the y-axis of plots."
    ),
    "plot_seaborn_style": (
        "darkgrid",
        "Defines the plot background/grid.\n"
        + "Options: 'whitegrid', 'darkgrid', 'white' or 'dark'."
    ),
    "plot_seaborn_palette": (
        "deep6",
        "Default color palette of seaborn. Can also be a list of colors.\n"
        + "See: https://seaborn.pydata.org/generated/seaborn.color_palette.html"
    ),
    "plot_export_format": (
        "pdf",
        "File format supported by matplotlib for exporting plots."
    ),
    "table_export_format": (
        "csv",
        "Format for exporting tables, e.g. 'csv', 'excel', 'latex', 'json'...",
    ),
    "table_export_data": (
        "stats",
        "Which data to export: 'info', 'stats' or 'error_array'.",
    ),
    "table_export_transpose": (
        True,
        "Transpose tables for export."
    ),
    "save_traj_in_zip": (
        False,
        "Store backup trajectories in result zip files (increases size)."
    ),
    "logging_format": (
        "%(message)s",
        "Format string for the logging module (console only)."
    ),
    "logfile_enabled": (
        False,
        "Whether to write a logfile to the home folder."
    )
}
# yapf: enable

# without documentation
DEFAULT_SETTINGS_DICT = {k: v[0] for k, v in DEFAULT_SETTINGS_DICT_DOC.items()}



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
main executable for trajectory analysis
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import logging

logger = logging.getLogger(__name__)

SEP = "-" * 80


def parser():
    import argparse
    basic_desc = "trajectory analysis and manipulation tool"
    lic = "(c) michael.grupp@tum.de"
    shared_parser = argparse.ArgumentParser(add_help=False)
    algo_opts = shared_parser.add_argument_group("algorithm options")
    output_opts = shared_parser.add_argument_group("output options")
    usability_opts = shared_parser.add_argument_group("usability options")
    shared_parser.add_argument("-f", "--full_check",
                               help="run all checks and print all stats",
                               action="store_true")
    algo_opts.add_argument(
        "-a", "--align", help="alignment with Umeyama's method (no scale)"
        " - requires --ref", action="store_true")
    algo_opts.add_argument(
        "-s", "--correct_scale", help="scale correction with Umeyama's method"
        " - requires --ref", action="store_true")
    algo_opts.add_argument(
        "--n_to_align",
        help="the number of poses to use for Umeyama alignment, "
        "counted from the start (default: all)", default=-1, type=int)
    algo_opts.add_argument(
        "--sync",
        help="associate trajectories via matching timestamps - requires --ref",
        action="store_true")
    algo_opts.add_argument(
        "--transform_left", help="path to a .json file with a transformation"
        " to apply to the trajectories (left multiplicative)")
    algo_opts.add_argument(
        "--transform_right", help="path to a .json file with a transformation"
        " to apply to the trajectories (right_multiplicative)")
    algo_opts.add_argument("--invert_transform",
                           help="invert the transformation of the .json file",
                           action="store_true")
    algo_opts.add_argument(
        "--ref", help="trajectory that will be marked/used as the reference")
    algo_opts.add_argument(
        "--t_offset",
        help="add a constant timestamp offset (not adding to --ref trajectory)",
        default=0.0, type=float)
    algo_opts.add_argument(
        "--t_max_diff",
        help="maximum timestamp difference for data association", default=0.01,
        type=float)
    algo_opts.add_argument(
        "--merge", help="merge the trajectories in a single trajectory",
        action="store_true")
    output_opts.add_argument("-p", "--plot", help="show plot window",
                             action="store_true")
    output_opts.add_argument(
        "--plot_mode", help="the axes for  plot projection", default=None,
        choices=["xy", "xz", "yx", "yz", "zx", "zy", "xyz"])
    output_opts.add_argument("--save_plot", help="path to save plot",
                             default=None)
    output_opts.add_argument("--serialize_plot",
                             help="path to serialize plot (experimental)",
                             default=None)
    output_opts.add_argument("--save_as_tum",
                             help="save trajectories in TUM format (as *.tum)",
                             action="store_true")
    output_opts.add_argument("--save_as_kitti",
                             help="save poses in KITTI format (as *.kitti)",
                             action="store_true")
    output_opts.add_argument("--save_as_bag",
                             help="save trajectories in ROS bag as <date>.bag",
                             action="store_true")
    usability_opts.add_argument("--no_warnings",
                                help="no warnings requiring user confirmation",
                                action="store_true")
    usability_opts.add_argument("-v", "--verbose", help="verbose output",
                                action="store_true")
    usability_opts.add_argument("--silent", help="don't print any output",
                                action="store_true")
    usability_opts.add_argument(
        "--debug", help="verbose output with additional debug info",
        action="store_true")
    usability_opts.add_argument(
        "-c", "--config",
        help=".json file with parameters (priority over command line args)")

    main_parser = argparse.ArgumentParser(
        description="%s %s" % (basic_desc, lic))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True
    kitti_parser = sub_parsers.add_parser(
        "kitti",
        description="%s for KITTI pose files - %s" % (basic_desc, lic),
        parents=[shared_parser])
    kitti_parser.add_argument("pose_files", help="one or multiple pose files",
                              nargs='+')

    tum_parser = sub_parsers.add_parser(
        "tum",
        description="%s for TUM trajectory files - %s" % (basic_desc, lic),
        parents=[shared_parser])
    tum_parser.add_argument("traj_files",
                            help="one or multiple trajectory files", nargs='+')

    euroc_parser = sub_parsers.add_parser(
        "euroc",
        description="%s for EuRoC MAV .csv's - %s" % (basic_desc, lic),
        parents=[shared_parser])
    euroc_parser.add_argument(
        "state_gt_csv",
        help="<sequence>/mav0/state_groundtruth_estimate0/data.csv", nargs='+')

    bag_parser = sub_parsers.add_parser(
        "bag", description="%s for ROS bag files - %s" % (basic_desc, lic),
        parents=[shared_parser])
    bag_parser.add_argument("bag", help="ROS bag file")
    bag_parser.add_argument(
        "topics",
        help="multiple geometry_msgs/PoseStamped or nav_msgs/Odometry topics",
        nargs='*')
    bag_parser.add_argument(
        "--all_topics", help=
        "use all geometry_msgs/PoseStamped and nav_msgs/Odometry topics in the bag",
        action="store_true")
    return main_parser


def die(msg):
    import sys
    logger.error(msg)
    sys.exit(1)


def load_trajectories(args):
    from evo.tools import file_interface
    trajectories = {}
    ref_traj = None
    if args.subcommand == "tum":
        for traj_file in args.traj_files:
            if traj_file == args.ref:
                continue
            trajectories[traj_file] = file_interface.read_tum_trajectory_file(
                traj_file)
        if args.ref:
            ref_traj = file_interface.read_tum_trajectory_file(args.ref)
    elif args.subcommand == "kitti":
        for pose_file in args.pose_files:
            if pose_file == args.ref:
                continue
            trajectories[pose_file] = file_interface.read_kitti_poses_file(
                pose_file)
        if args.ref:
            ref_traj = file_interface.read_kitti_poses_file(args.ref)
    elif args.subcommand == "euroc":
        for csv_file in args.state_gt_csv:
            if csv_file == args.ref:
                continue
            else:
                trajectories[
                    csv_file] = file_interface.read_euroc_csv_trajectory(
                        csv_file)
        if args.ref:
            ref_traj = file_interface.read_euroc_csv_trajectory(args.ref)
    elif args.subcommand == "bag":
        if not (args.topics or args.all_topics):
            die("No topics used - specify topics or set --all_topics.")
        import rosbag
        bag = rosbag.Bag(args.bag)
        try:
            if args.all_topics:
                topic_info = bag.get_type_and_topic_info()
                topics = sorted([
                    t for t in topic_info[1].keys() if topic_info[1][t][0] in
                    {"geometry_msgs/PoseStamped", "nav_msgs/Odometry"}
                    and t != args.ref
                ])
                if len(topics) == 0:
                    die("No geometry_msgs/PoseStamped or nav_msgs/Odometry "
                        "topics found!")
            else:
                topics = args.topics
            for topic in topics:
                trajectories[topic] = file_interface.read_bag_trajectory(
                    bag, topic)
            if args.ref:
                ref_traj = file_interface.read_bag_trajectory(bag, args.ref)
        finally:
            bag.close()
    return trajectories, ref_traj


# TODO refactor
def print_traj_info(name, traj, verbose=False, full_check=False):
    import os
    from evo.core import trajectory

    logger.info(SEP)
    logger.info("name:\t" + os.path.splitext(os.path.basename(name))[0])

    if verbose or full_check:
        infos = traj.get_infos()
        info_str = ""
        for info, value in sorted(infos.items()):
            info_str += "\n\t" + info + "\t" + str(value)
        logger.info("infos:" + info_str)
        if full_check:
            passed, details = traj.check()
            check_str = ""
            for test, result in sorted(details.items()):
                check_str += "\n\t" + test + "\t" + result
            logger.info("checks:" + check_str)
            stat_str = ""
            try:
                stats = traj.get_statistics()
                for stat, value in sorted(stats.items()):
                    if isinstance(value, float):
                        stat_str += "\n\t" + stat + "\t" + "{0:.6f}".format(
                            value)
                    else:
                        stat_str += value
            except trajectory.TrajectoryException as e:
                stat_str += "\n\terror - " + str(e)
            logger.info("stats:" + stat_str)
    else:
        logger.info("infos:\t" + str(traj))


def run(args):
    import os
    import sys

    import numpy as np

    import evo.core.lie_algebra as lie
    from evo.core import trajectory
    from evo.core.trajectory import PoseTrajectory3D
    from evo.tools import file_interface, log
    from evo.tools.settings import SETTINGS

    log.configure_logging(verbose=args.verbose, silent=args.silent,
                          debug=args.debug)
    if args.debug:
        import pprint
        logger.debug("main_parser config:\n" + pprint.pformat(
            {arg: getattr(args, arg)
             for arg in vars(args)}) + "\n")
    logger.debug(SEP)

    trajectories, ref_traj = load_trajectories(args)

    if args.merge:
        if args.subcommand == "kitti":
            die("Can't merge KITTI files.")
        if len(trajectories) == 0:
            die("No trajectories to merge (excluding --ref).")
        trajectories = {
            "merged_trajectory": trajectory.merge(trajectories.values())
        }

    if args.transform_left or args.transform_right:
        tf_type = "left" if args.transform_left else "right"
        tf_path = args.transform_left \
                if args.transform_left else args.transform_right
        transform = file_interface.load_transform_json(tf_path)
        logger.debug(SEP)
        if not lie.is_se3(transform):
            logger.warning("Not a valid SE(3) transformation!")
        if args.invert_transform:
            transform = lie.se3_inverse(transform)
        logger.debug("Applying a {}-multiplicative transformation:\n{}".format(
            tf_type, transform))
        for traj in trajectories.values():
            traj.transform(transform, right_mul=args.transform_right)

    if args.t_offset:
        logger.debug(SEP)
        for name, traj in trajectories.items():
            if type(traj) is trajectory.PosePath3D:
                die("{} doesn't have timestamps - can't add time offset.".
                    format(name))
            logger.info("Adding time offset to {}: {} (s)".format(
                name, args.t_offset))
            traj.timestamps += args.t_offset

    if args.sync or args.align or args.correct_scale:
        from evo.core import sync
        if not args.ref:
            logger.debug(SEP)
            die("Can't align or sync without a reference! (--ref)  *grunt*")
        for name, traj in trajectories.items():
            if args.subcommand == "kitti":
                ref_traj_tmp = ref_traj
            else:
                logger.debug(SEP)
                ref_traj_tmp, trajectories[name] = sync.associate_trajectories(
                    ref_traj, traj, max_diff=args.t_max_diff,
                    first_name="reference", snd_name=name)
            if args.align or args.correct_scale:
                logger.debug(SEP)
                logger.debug("Aligning {} to reference.".format(name))
                trajectories[name] = trajectory.align_trajectory(
                    trajectories[name], ref_traj_tmp,
                    correct_scale=args.correct_scale,
                    correct_only_scale=args.correct_scale and not args.align,
                    n=args.n_to_align)

    for name, traj in trajectories.items():
        print_traj_info(name, traj, args.verbose, args.full_check)
    if args.ref:
        print_traj_info(args.ref, ref_traj, args.verbose, args.full_check)

    if args.plot or args.save_plot or args.serialize_plot:
        from evo.tools.plot import PlotMode
        plot_mode = PlotMode.xyz if not args.plot_mode else PlotMode[args.
                                                                     plot_mode]
        import numpy as np
        from evo.tools import plot
        import matplotlib.pyplot as plt
        import matplotlib.cm as cm

        plot_collection = plot.PlotCollection("evo_traj - trajectory plot")
        fig_xyz, axarr_xyz = plt.subplots(3, sharex="col",
                                          figsize=tuple(SETTINGS.plot_figsize))
        fig_rpy, axarr_rpy = plt.subplots(3, sharex="col",
                                          figsize=tuple(SETTINGS.plot_figsize))
        fig_traj = plt.figure(figsize=tuple(SETTINGS.plot_figsize))
        ax_traj = plot.prepare_axis(fig_traj, plot_mode)

        if args.ref:
            short_traj_name = os.path.splitext(os.path.basename(args.ref))[0]
            if SETTINGS.plot_usetex:
                short_traj_name = short_traj_name.replace("_", "\\_")
            plot.traj(ax_traj, plot_mode, ref_traj, '--', 'grey',
                      short_traj_name, alpha=0 if SETTINGS.plot_hideref else 1)
            plot.traj_xyz(axarr_xyz, ref_traj, '--', 'grey', short_traj_name,
                          alpha=0 if SETTINGS.plot_hideref else 1)
            plot.traj_rpy(axarr_rpy, ref_traj, '--', 'grey', short_traj_name,
                          alpha=0 if SETTINGS.plot_hideref else 1)

        cmap_colors = None
        if SETTINGS.plot_multi_cmap.lower() != "none":
            cmap = getattr(cm, SETTINGS.plot_multi_cmap)
            cmap_colors = iter(cmap(np.linspace(0, 1, len(trajectories))))

        for name, traj in trajectories.items():
            if cmap_colors is None:
                color = next(ax_traj._get_lines.prop_cycler)['color']
            else:
                color = next(cmap_colors)
            short_traj_name = os.path.splitext(os.path.basename(name))[0]
            if SETTINGS.plot_usetex:
                short_traj_name = short_traj_name.replace("_", "\\_")
            plot.traj(ax_traj, plot_mode, traj, '-', color, short_traj_name)
            if args.ref and isinstance(ref_traj, trajectory.PoseTrajectory3D):
                start_time = ref_traj.timestamps[0]
            else:
                start_time = None
            plot.traj_xyz(axarr_xyz, traj, '-', color, short_traj_name,
                          start_timestamp=start_time)
            plot.traj_rpy(axarr_rpy, traj, '-', color, short_traj_name,
                          start_timestamp=start_time)

        plot_collection.add_figure("trajectories", fig_traj)
        plot_collection.add_figure("xyz_view", fig_xyz)
        plot_collection.add_figure("rpy_view", fig_rpy)
        if args.plot:
            plot_collection.show()
        if args.save_plot:
            logger.info(SEP)
            plot_collection.export(args.save_plot,
                                   confirm_overwrite=not args.no_warnings)
        if args.serialize_plot:
            logger.info(SEP)
            plot_collection.serialize(args.serialize_plot,
                                      confirm_overwrite=not args.no_warnings)

    if args.save_as_tum:
        logger.info(SEP)
        for name, traj in trajectories.items():
            dest = os.path.splitext(os.path.basename(name))[0] + ".tum"
            file_interface.write_tum_trajectory_file(
                dest, traj, confirm_overwrite=not args.no_warnings)
        if args.ref:
            dest = os.path.splitext(os.path.basename(args.ref))[0] + ".tum"
            file_interface.write_tum_trajectory_file(
                dest, ref_traj, confirm_overwrite=not args.no_warnings)
    if args.save_as_kitti:
        logger.info(SEP)
        for name, traj in trajectories.items():
            dest = os.path.splitext(os.path.basename(name))[0] + ".kitti"
            file_interface.write_kitti_poses_file(
                dest, traj, confirm_overwrite=not args.no_warnings)
        if args.ref:
            dest = os.path.splitext(os.path.basename(args.ref))[0] + ".kitti"
            file_interface.write_kitti_poses_file(
                dest, ref_traj, confirm_overwrite=not args.no_warnings)
    if args.save_as_bag:
        import datetime
        import rosbag
        dest_bag_path = str(
            datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")) + ".bag"
        logger.info(SEP)
        logger.info("Saving trajectories to " + dest_bag_path + "...")
        bag = rosbag.Bag(dest_bag_path, 'w')
        try:
            for name, traj in trajectories.items():
                dest_topic = os.path.splitext(os.path.basename(name))[0]
                frame_id = traj.meta[
                    "frame_id"] if "frame_id" in traj.meta else ""
                file_interface.write_bag_trajectory(bag, traj, dest_topic,
                                                    frame_id)
            if args.ref:
                dest_topic = os.path.splitext(os.path.basename(args.ref))[0]
                frame_id = ref_traj.meta[
                    "frame_id"] if "frame_id" in ref_traj.meta else ""
                file_interface.write_bag_trajectory(bag, ref_traj, dest_topic,
                                                    frame_id)
        finally:
            bag.close()


if __name__ == '__main__':
    from evo import entry_points
    entry_points.traj()
